import xarray as xr
import matplotlib.pyplot as plt
import numpy as np
import logging


logger = logging.getLogger(__name__)


def plot_debug_poe(raw_fcst, cal_fcst, latlon, geogrid=None, ptiles=None, climo=None,
                   var_info=None):
    logger.debug(f"Creating a POE ")
    if var_info is None:
        raise ValueError('Must include var_info parameter, see API docs')
    if ptiles is None:
        ptiles = [1, 2, 5, 10, 15, 20, 25, 33, 40, 50, 60, 67, 75, 80, 85, 90, 95, 98, 99]
    num_members = raw_fcst.shape[0]
    num_lats = len(geogrid.lats)
    num_lons = len(geogrid.lons)
    plot_lat = latlon[0]
    plot_lon = latlon[1] if latlon[1] >= 0 else 360 + latlon[1]

    raw_fcst_da = xr.DataArray(raw_fcst.reshape(num_members, num_lats, num_lons),
                               dims=['member', 'lat', 'lon'],
                               coords={'member': list(range(num_members)), 'lat': geogrid.lats,
                                       'lon': geogrid.lons})
    cal_fcst_da = xr.DataArray(cal_fcst.reshape(len(ptiles), num_lats, num_lons),
                               dims=['ptile', 'lat', 'lon'],
                               coords={'ptile': ptiles, 'lat': geogrid.lats, 'lon': geogrid.lons})

    climo_da = xr.DataArray(climo.reshape(len(ptiles), num_lats, num_lons),
                               dims=['ptile', 'lat', 'lon'],
                               coords={'ptile': ptiles, 'lat': geogrid.lats, 'lon': geogrid.lons})

    fig, ax = plt.subplots()
    ax.grid(True)
    ax.set_title(f"{var_info['name'].capitalize()} POE Forecast at ({plot_lat}N, {plot_lon}E)")
    ax.plot(np.exp(np.flip(np.sort(raw_fcst_da.loc[dict(lat=plot_lat, lon=plot_lon)])),),
            raw_fcst_da.member / (len(raw_fcst_da.member) - 1),
            label='Raw fcst')
    ax.set_xlabel(f"{var_info['name']}")
    ax.plot(climo_da.loc[dict(lat=plot_lat, lon=plot_lon)],
            cal_fcst_da.loc[dict(lat=plot_lat, lon=plot_lon)],
            label='Calibrated fcst')
    ax.set_ylabel('Forecast Probability')
    ax.set_ylim([0, 1])
    legend = ax.legend(loc='upper right')
    for i in range(len(climo_da.loc[dict(lat=plot_lat, lon=plot_lon)])):
        x = float(climo_da.loc[dict(lat=plot_lat, lon=plot_lon)][i])
        p = int(climo_da.loc[dict(lat=plot_lat, lon=plot_lon)].ptile[i])
        if p == 50:
            linewidth = 1
            linestyle = '-'
        else:
            linewidth = 0.3
            linestyle = '--'
        plt.axvline(x, linewidth=linewidth, linestyle=linestyle, color='red')
        ax.annotate(str(p), (x, -0.005), horizontalalignment='center', verticalalignment='top',
                    fontsize=6, fontweight='light', color='red', annotation_clip=False)

    # plt.tight_layout()
    plt.savefig('cal-poe.png', dpi=300)


if __name__ == '__main__':
    import pickle

    with open('/export/cpc-lw-mcharles/mcharles/pycharm-deployments/mpp-driver/mpp_driver/data.pkl', 'rb') as file:
        raw_fcst, cal_fcst, debug_poe_latlon, geogrid, climo = pickle.load(file)
        debug_poe_latlon = tuple([float(val) for val in debug_poe_latlon])

    plot_debug_poe(raw_fcst, cal_fcst, debug_poe_latlon, geogrid=geogrid, climo=climo,
                   var_info={'name': 'precip'})



# Third-party
import numpy as np
from scipy.stats import norm

# CPC
import cpc.geogrids

# This application
from .plot_poe_charts import plot_debug_poe


class StatsError(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


def ensemble_regression(raw_fcst, stats, method='ensemble', ens_size_correction=False,
                        ptiles=list([1, 2, 5, 10, 15, 20, 25, 33, 40, 50, 60, 67, 75, 80, 85, 90, 95, 98, 99]),
                        debug=False, variable='tmean', debug_poe_latlon=None, climo=None):
    # ----------------------------------------------------------------------------------------------
    # Check stats dict for all required stats
    #
    required_stats = ['cov', 'es', 'xm', 'xv', 'ym', 'yv', 'num_stats_members',
                      'num_fcst_members', 'num_years']
    for required_stat in required_stats:
        if required_stat not in stats.keys():
            raise StatsError('Stat {} not found in stats. Required stats are: {}'.format(
                required_stat, required_stats))

    # ----------------------------------------------------------------------------------------------
    # Set some config options
    #
    xv_min = 0.1    # min xv allowed
    rxy_min = 0.05  # min rxy allowed
    rxy_max = 0.90  # max rxy allowed

    # ----------------------------------------------------------------------------------------------
    # Extract fields from stats dict
    #
    cov = stats['cov']
    es = stats['es']
    xm = stats['xm']
    xv = stats['xv']
    ym = stats['ym']
    yv = stats['yv']
    num_stats_members = stats['num_stats_members']
    num_fcst_members = stats['num_fcst_members']
    num_years = stats['num_years']

    # ----------------------------------------------------------------------------------------------
    # Set bounds for some stats
    #
    # Set lower bound for xv
    xv = np.where(xv < xv_min, xv_min, xv)
    # Set lower and upper bounds for yv
    yv_min = 0.1 * xv  # min yv allowed
    yv = np.where(yv < yv_min, yv_min, yv)
    yv_max = xv  # signal cannot be greater than obs variance
    yv = np.where(yv > yv_max, yv_max, yv)
    # Set lower and upper bounds for cov
    cov_min = 0  # ignore negative covariances and corelations
    cov = np.where(cov < cov_min, cov_min, cov)
    cov_max = xv  # Max covariance equal to variance of obs
    cov = np.where(cov > cov_max, cov_max, cov)

    # ----------------------------------------------------------------------------------------------
    # Adjust ym to look more like the 30-year climo
    #
    if 'adjust_model_climo' in stats:
        ym += stats['climo_mean'] - xm

    # ----------------------------------------------------------------------------------------------
    # Calculate correlation of obs and ens mean
    #
    rxy = cov / np.sqrt(xv * yv)
    # Limit values of rxy
    rxy = np.where(rxy > rxy_max, rxy_max, rxy)

    # ----------------------------------------------------------------------------------------------
    # Correct stats, compensating for the difference between the number of members used to
    # correct the stats, and the number of members in the real time forecast
    #
    if ens_size_correction:
        es = es * (num_stats_members / num_fcst_members) * \
             (num_fcst_members - 1) / (num_stats_members - 1)
        yv_uncorrected = yv
        yv = yv - (num_fcst_members - num_stats_members)
        rxy = rxy * np.sqrt(yv / yv_uncorrected)

    # ----------------------------------------------------------------------------------------------
    # Calculate correlation of obs and best member
    #
    rbest = rxy * np.sqrt(1 + es / yv)

    # ----------------------------------------------------------------------------------------------
    # Calculate fcst anomalies and mean
    #
    y_anom = raw_fcst - ym
    y_anom_mean = np.nanmean(y_anom, axis=0)

    # ----------------------------------------------------------------------------------------------
    # Calculate correction for over-dispersive model
    #
    with np.errstate(divide='ignore', invalid='ignore'):
        k = np.sqrt(yv / es * (num_fcst_members - 1) / num_fcst_members * (1 / rxy ** 2 - 1))
    k = np.where(k > 1, 1, k)  # keep kn <= 1
    k = np.where(rbest > 1, k, 1)  # set to 1 if rbest not > 1

    # ----------------------------------------------------------------------------------------------
    # Adjust fcst members to correct for over-dispersion
    #
    y_anom = y_anom_mean + k * (y_anom - y_anom_mean)

    # ----------------------------------------------------------------------------------------------
    # Make adjustments for cases where correlations are low (in this case, use climo)
    #
    rxy = np.where(rxy < rxy_min, rxy_min, rxy)

    # ----------------------------------------------------------------------------------------------
    # Calculate regression coefficient and error of best member
    #
    a1 = rxy * np.sqrt(xv / yv)
    ebest = np.sqrt(num_years / (num_years - 2) * xv * (1 - rxy**2 * (1 + k**2 * es/yv)))
    emean = np.sqrt(num_years / (num_years - 2) * xv * (1 - rxy**2))
    # Set lower bound for ebest - if ebest is zero, then the scale parameter in norm.cdf will be
    # zero, resulting in the cdf being NaN - set to a very small number instead of zero
    ebest = np.where(ebest == 0, 0.00001, ebest)

    # ----------------------------------------------------------------------------------------------
    # Correct each member
    #
    norm_ptiles = norm.ppf(np.array(ptiles) / 100)
    POE_ens_mean = np.full((len(norm_ptiles), raw_fcst.shape[1]), np.nan)
    POE_ens = np.full((len(norm_ptiles), raw_fcst.shape[0], raw_fcst.shape[1]), np.nan)
    for p, norm_ptile in enumerate(norm_ptiles):
        for m in range(raw_fcst.shape[0]):
            with np.errstate(divide='ignore', invalid='ignore'):
                POE_ens[p, m] = 1 - norm.cdf(norm_ptile, a1 * y_anom[m] / np.sqrt(xv),
                                             ebest / np.sqrt(xv))
    POE_ens_mean = np.nanmean(POE_ens, axis=1)

    # ----------------------------------------------------------------------------------------------
    # Plot stats for debugging
    #
    if debug:
        from cpc.geogrids import Geogrid
        from cpc.geoplot import Geomap, Geofield

        geogrid = Geogrid('1deg-global')

        levels = {
            'tmean': {
                'a1': np.arange(0.1, 1.3, 0.1),
                'ebest': np.arange(0.5, 6.5, 0.5),
                'emean': np.arange(0.5, 7.5, 0.5),
                'k': np.arange(0.1, 1.1, 0.1),
                'rxy': np.arange(0.1, 1.1, 0.1),
                'es': np.arange(3, 28, 3),
                'yv': np.arange(5, 56, 5),
                'xv': np.arange(5, 56, 5),
                'rbest': np.arange(0, 1.3, 0.1),
                'y_anom_mean': 'auto',
                'y_anom_single': np.arange(-10, 16, 1),
            },
            'precip': {
                'a1': np.arange(0.1, 1.3, 0.1),
                'ebest': np.arange(0.5, 3, 0.2),
                'emean': np.arange(0.5, 3, 0.2),
                'k': np.arange(0.1, 1.1, 0.1),
                'rxy': np.arange(0.1, 1.1, 0.1),
                'es': np.arange(1, 11, 1),
                'yv': np.arange(1, 11, 1),
                'xv': np.arange(1, 11, 1),
                'rbest': np.arange(0, 1.3, 0.1),
                'y_anom_mean': 'auto',
                'y_anom_single': 'auto',
            }
        }

        for stat in ['a1', 'ebest', 'emean', 'k', 'rxy', 'es', 'yv', 'rbest', 'xv', 'y_anom_mean']:
            with Geomap() as geomap:
                geofield = Geofield(locals()[stat], geogrid, levels=levels[variable][stat])
                geomap.plot(geofield)
                geomap.save('{}.png'.format(stat), dpi=200)

    if debug_poe_latlon:
        plot_debug_poe(raw_fcst, POE_ens_mean, debug_poe_latlon, geogrid, climo=climo,
                       var_info={'name': variable})


    # ----------------------------------------------------------------------------------------------
    # Return total POE
    #
    return POE_ens_mean



#!/usr/bin/env python
# -*- coding: utf-8 -*-


try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup
from setuptools import find_packages

with open('README.md') as readme_file:
    readme = readme_file.read()

with open('CHANGELOG.md') as history_file:
    history = history_file.read()

requirements = [
    'scipy',
    'numpy',
    'cpc.geofiles',
    'cpc.stats'
]

setup(
    name="cpc.mpp",
    version='v0.2.7',
    description="Model post processing package",
    long_description=readme + '\n\n' + history,
    author="Mike Charles",
    author_email='mike.charles@noaa.gov',
    url="https://github.com/mikecharles/cpc.mpp",
    packages=find_packages(),
    namespace_packages=['cpc'],
    include_package_data=True,
    install_requires=requirements,
    license="CC",
    zip_safe=False,
    classifiers=[
        'Development Status :: 1 - Planning',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent',
        'Topic :: Scientific/Engineering :: Atmospheric Science',
        'License :: CC0 1.0 Universal (CC0 1.0) Public Domain Dedication',
    ],
)



import numpy as np
from scipy import stats
import logging
import matplotlib.pyplot as plt


def _plot_correction(A_cdf, B_cdf, A_dist, B_dist, quantile_correction, C,
                     C_corrected, quantiles):
    # --------------------------------------------------------------------------
    # Plot correction in CDF space
    #
    plt.figure()
    # Figure settings
    linewidth = 2
    markeredgewidth = 2
    # Plot sample CDFs
    plt.plot(quantiles[1:], A_cdf, 'x', c='r', markersize=8,
             markeredgewidth=markeredgewidth, label='Fcst')
    plt.plot(quantiles[1:], B_cdf, '+', c='g', markersize=8,
             markeredgewidth=markeredgewidth, label='Obs')
    # Plot fitted CDFs
    plt.plot(quantiles[1:], A_dist.cdf(quantiles[1:]), '-', color='r',
             linewidth=linewidth, label='Gamma approx. of Fcst')
    plt.plot(quantiles[1:], B_dist.cdf(quantiles[1:]), '-', color='g',
             linewidth=linewidth, label='Gamma approx. of Obs')
    # Plot arrows indicating CDF correction
    for q in range(len(quantiles[1:])):
        if ~np.isinf(quantile_correction[q]) and quantile_correction[q] != 0:
            print('Correction at {}: {}'.format(quantiles[q],
                                                quantile_correction[q]))
            plt.arrow(
                quantiles[q + 1], A_dist.cdf(quantiles[q + 1]),
                quantile_correction[q], 0, length_includes_head=True,
                head_width=0.03, head_length=2, edgecolor='k', facecolor='k',
                overhang=1
            )
    # Add legend, title, and labels
    plt.legend(loc='lower right')
    plt.title('CDF Matching - Making Fcst Look Like Obs')
    plt.xlabel('Quantile Values (units of sample)')
    plt.ylabel('Probability Density')
    # plt.show()
    plt.savefig('stats1.png', dpi=200, bbox_inches='tight')
    # --------------------------------------------------------------------------
    # Plot correction in sample space
    #
    plt.figure()
    # Figure settings
    linewidth = 2
    markeredgewidth = 2
    # Plot C and C_corrected
    plt.plot(sorted(C), range(len(C)), 'r', linewidth=linewidth,
             label='Sample C')
    plt.plot(sorted(C_corrected), range(len(C)), 'g', linewidth=linewidth,
             label='Sample C (corrected)')
    # Add legend, title, and labels
    plt.legend(loc='lower right')
    plt.title('Sample C Before and After Correction')
    plt.xlabel('Sample value')
    plt.ylabel('Sample index')
    plt.savefig('stats2.png', dpi=200, bbox_inches='tight')


def cdf_correction(past_fcst_data, past_obs_data, curr_fcst_data, quantiles,
                   plot_point=None):
    # Make input NumPy arrays copies, so as not to modify the original arrays
    # in the calling function
    past_fcst_data = past_fcst_data.copy()
    past_obs_data = past_obs_data.copy()
    curr_fcst_data = curr_fcst_data.copy()
    # Initialize an array to store the corrected forecast data
    curr_fcst_data_corrected = np.empty(curr_fcst_data.shape) * np.nan
    # Flatten forecast array so all ensemble members and days are pooled
    # together. For example, for 45 days of past 1-deg forecasts, the array
    # shape would be (45 x 21 x 65160). We want to to flatten that to be
    # (945 x 65160).
    if past_fcst_data.ndim == 3:
        past_fcst_data = past_fcst_data.reshape(
            past_fcst_data.shape[0] * past_fcst_data.shape[1],
            past_fcst_data.shape[2])
    # Loop over all grid points
    for g in range(past_fcst_data.shape[1]):
        # ----------------------------------------------------------------------
        # QC data
        #
        qc_pcnt_data_req_at_gdpt = 90
        pcnt_data_found = 100 * np.count_nonzero(~np.isnan(past_obs_data[:, g])
                                                 ) / past_obs_data.shape[0]
        if pcnt_data_found < qc_pcnt_data_req_at_gdpt:
            continue
        pcnt_data_found = 100 * np.count_nonzero(~np.isnan(past_fcst_data[:, g])
                                                 ) / past_fcst_data.shape[0]
        if pcnt_data_found < qc_pcnt_data_req_at_gdpt:
            continue

        # ----------------------------------------------------------------------
        # Edge cases
        #
        # If obs are all zeros, set corrected forecast to zero
        if np.all(past_obs_data[:, g] == 0):
            curr_fcst_data_corrected[:, g] = 0
            continue
        # If obs are all the same, set the mean of the corrected forecasts to
        # that value
        if np.unique(past_obs_data[:, g]).size == 1:
            curr_fcst_data_corrected[:, g] = curr_fcst_data[:, g] + \
                (past_obs_data[0, g] - np.nanmean(past_fcst_data[:, g]))
            continue
        # If obs are all NaNs, set the corrected forecast to NaNs (can't
        # correct)
        if np.all(np.isnan(past_obs_data[:, g])):
            curr_fcst_data_corrected[:, g] = np.nan
            continue
        # If fcsts are all zero, set corrected forecast to zero
        if np.all(past_fcst_data[:, g] == 0):
            curr_fcst_data_corrected[:, g] = 0
            continue

        # ----------------------------------------------------------------------
        # Add a tiny amount so values are > 0
        #
        past_obs_data[:, g] = past_obs_data[:, g] + 0.001
        past_fcst_data[:, g] = past_fcst_data[:, g] + 0.001

        # ----------------------------------------------------------------------
        # Calculate PDFs and CDFs from the forecast and observation samples
        #
        past_fcst_pdf, edges = np.histogram(past_fcst_data[:, g], bins=quantiles)
        past_fcst_cdf = np.cumsum(past_fcst_pdf) / past_fcst_data.shape[0]
        past_obs_pdf, edges = np.histogram(past_obs_data[:, g], bins=quantiles)
        past_obs_cdf = np.cumsum(past_obs_pdf) / past_obs_data.shape[0]
        # Fit a gamma distribution
        try:
            past_fcst_dist = stats.gamma(
                *stats.gamma.fit(
                    past_fcst_data[np.nonzero(past_fcst_data[:, g]), g], floc=0
                )
            )
            past_obs_dist = stats.gamma(
                *stats.gamma.fit(
                    past_obs_data[np.nonzero(past_obs_data[:, g]), g], floc=0
                )
            )
        # If a ValueError is raised during curve-fitting, set the corrected
        # fcst data to np.nan
        except ValueError:
            curr_fcst_data_corrected[:, g] = np.nan
            continue
        # Calculate corrected quantiles
        corrected_quantiles = past_obs_dist.ppf(past_fcst_dist.cdf(quantiles[1:]))
        quantile_correction = corrected_quantiles - quantiles[1:]
        curr_fcst_data_corrected[:, g] = past_obs_dist.ppf(past_fcst_dist.cdf(
            curr_fcst_data[:, g]))
        # ----------------------------------------------------------------------
        # Plot
        #
        if plot_point and g == plot_point:
            try:
                _plot_correction(
                    past_fcst_cdf, past_obs_cdf, past_fcst_dist, past_obs_dist,
                    quantile_correction, curr_fcst_data[:, g],
                    curr_fcst_data_corrected[:, g], quantiles
                )
            except:
                pass

    # --------------------------------------------------------------------------
    # Fix any remaining issues with the corrected forecast
    #
    # Set all negative values to 0
    with np.errstate(invalid='ignore'):
        curr_fcst_data_corrected = np.where(curr_fcst_data_corrected < 0, 0,
                                        curr_fcst_data_corrected)

    return curr_fcst_data_corrected


def bias_correction(past_fcst_data, past_obs_data, curr_fcst_data, weights=None):
    # If past_fcst_data is 1-d, add a 2nd dimension, since np.average() below expects it to be 2-d. It would be 2-d
    # if correction_type was set to st-bc (grid points x days)
    if past_fcst_data.ndim == 1:
        past_fcst_data = np.expand_dims(past_fcst_data, 0)
    elif past_fcst_data.ndim == 2:
        pass
    else:
        raise ValueError(f'past_fcst_data has a dimension of {past_fcst_data.ndim}, but needs to have a dimension of '
                         f'1 or 2')

    # Calculate past bias
    bias = np.average(
        np.ma.masked_invalid(past_fcst_data) - np.ma.masked_invalid(past_obs_data), axis=0, weights=weights
    )

    # Correct current fcst, given the past bias
    corrected_curr_fcst_data = curr_fcst_data - bias

    return corrected_curr_fcst_data
